from aws_cdk.aws_apigatewayv2 import CfnAuthorizer, CfnApi
from aws_cdk.core import Stack
from b_cfn_custom_userpool_authorizer.user_pool_config import UserPoolConfig

from b_cfn_custom_userpool_authorizer.user_pool_custom_authorizer_function import AuthorizerFunction


class UserPoolCustomAuthorizer(CfnAuthorizer):
    def __init__(
            self,
            scope: Stack,
            name: str,
            api: CfnApi,
            user_pool_config: UserPoolConfig,
    ) -> None:
        """
        Constructor.

        :param scope: CloudFormation stack.
        :param name: Name of the custom authorizer e.g. "MyCoolAuthorizer".
        :param api: Parent API for which we are creating the authorizer.
        """
        lambda_function = AuthorizerFunction(
            scope=scope,
            name=f'{name}Function',
            user_pool_config=user_pool_config
        )

        # Constructed by reading this documentation:
        # https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-apigatewayv2-authorizer.html
        super().__init__(
            scope=scope,
            id=name,
            name=name,
            api_id=api.ref,
            authorizer_payload_format_version='2.0',
            authorizer_result_ttl_in_seconds=60,
            authorizer_type='REQUEST',
            authorizer_uri=(
                f'arn:aws:apigateway:{scope.region}:'
                f'lambda:path/2015-03-31/functions/arn:'
                f'aws:lambda:{scope.region}:{scope.account}:'
                f'function:{lambda_function}/invocations'
            ),
            identity_source=[],
        )



class PolicyDocument:
    def __init__(self):
        region = 'region'
        account_id = 'account'
        api_id = 'id'


        document = {
            "policyDocument": {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Action": "execute-api:Invoke",
                        "Resource": [
                            f"arn:aws:execute-api:{region}:{account_id}:{api_id}/*/*"
                        ],
                        "Effect": "Allow"
                    }
                ]
            }
        }



# Copyright 2017-2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not use this file
# except in compliance with the License. A copy of the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on an "AS IS"
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under the License.

import json
import time
import urllib.request
from jose import jwk, jwt
from jose.utils import base64url_decode

region = 'ap-southeast-2'
userpool_id = 'ap-southeast-2_xxxxxxxxx'
app_client_id = '<ENTER APP CLIENT ID HERE>'
keys_url = 'https://cognito-idp.{}.amazonaws.com/{}/.well-known/jwks.json'.format(region, userpool_id)
# instead of re-downloading the public keys every time
# we download them only on cold start
# https://aws.amazon.com/blogs/compute/container-reuse-in-lambda/
with urllib.request.urlopen(keys_url) as f:
    response = f.read()
keys = json.loads(response.decode('utf-8'))['keys']


def handler(event, context):
    token = event['token']
    # get the kid from the headers prior to verification
    headers = jwt.get_unverified_headers(token)
    kid = headers['kid']
    # search for the kid in the downloaded public keys
    key_index = -1
    for i in range(len(keys)):
        if kid == keys[i]['kid']:
            key_index = i
            break
    if key_index == -1:
        print('Public key not found in jwks.json')
        return False
    # construct the public key
    public_key = jwk.construct(keys[key_index])
    # get the last two sections of the token,
    # message and signature (encoded in base64)
    message, encoded_signature = str(token).rsplit('.', 1)
    # decode the signature
    decoded_signature = base64url_decode(encoded_signature.encode('utf-8'))
    # verify the signature
    if not public_key.verify(message.encode("utf8"), decoded_signature):
        print('Signature verification failed')
        return False
    print('Signature successfully verified')
    # since we passed the verification, we can now safely
    # use the unverified claims
    claims = jwt.get_unverified_claims(token)
    # additionally we can verify the token expiration
    if time.time() > claims['exp']:
        print('Token is expired')
        return False
    # and the Audience  (use claims['client_id'] if verifying an access token)
    if claims['aud'] != app_client_id:
        print('Token was not issued for this audience')
        return False
    # now we can use the claims
    print(claims)
    return claims



from typing import Optional, Dict


class UserPoolConfig:
    def __init__(
            self,
            user_pool_id: Optional[str] = None,
            user_pool_client_id: Optional[str] = None,
            user_pool_id_ssm_key: Optional[str] = None,
            user_pool_client_id_ssm_key: Optional[str] = None
    ) -> None:
        self.user_pool_id = user_pool_id
        self.user_pool_client_id = user_pool_client_id
        self.id_specified = self.user_pool_id and self.user_pool_client_id

        self.user_pool_id_ssm_key = user_pool_id_ssm_key
        self.user_pool_client_id_ssm_key = user_pool_client_id_ssm_key
        self.ssm_specified = self.user_pool_id_ssm_key and self.user_pool_client_id_ssm_key

        if (not self.id_specified) and (not self.ssm_specified):
            raise ValueError('Specify either resource ids or ssm keys from which ids will be retrieved.')

        if self.id_specified and self.ssm_specified:
            raise ValueError('You can not specify both ids and ssm keys.')

    def to_dict(self) -> Dict[str, str]:
        if self.id_specified:
            return {
                'USER_POOL_ID': self.user_pool_id,
                'USER_POOL_CLIENT_ID': self.user_pool_client_id,
            }

        if self.ssm_specified:
            return {
                'USER_POOL_ID_SSM_KEY': self.user_pool_id_ssm_key,
                'USER_POOL_CLIENT_ID_SSM_KEY': self.user_pool_client_id_ssm_key,
            }

        raise Exception('Unexpected state.')
